import os

import pytest

from any_agent.tools import search_tavily


@pytest.mark.skipif(
    not os.getenv("TAVILY_API_KEY"),
    reason="TAVILY_API_KEY not set; skipping Tavily integration test!",
)
def test_search_tavily_real() -> None:
    query = "Who is Leo Messi?"
    result = search_tavily(query)
    # Expect at least a title or content in the result
    assert "Leo Messi".lower() in result.lower() or "No results found." not in result



from collections.abc import Mapping
from datetime import timedelta
from typing import TYPE_CHECKING, Any

from litellm.cost_calculator import cost_per_token
from pydantic import BaseModel, ConfigDict, Field

from any_agent.config import AgentFramework
from any_agent.logging import logger

from .otel_types import (
    AttributeValue,
    Event,
    Link,
    Resource,
    SpanContext,
    SpanKind,
    Status,
)

if TYPE_CHECKING:
    from opentelemetry.sdk.trace import ReadableSpan


class CountInfo(BaseModel):
    """Token Count information."""

    token_count_prompt: int
    token_count_completion: int

    model_config = ConfigDict(extra="forbid")


class CostInfo(BaseModel):
    """Cost information."""

    cost_prompt: float
    cost_completion: float

    model_config = ConfigDict(extra="forbid")


class TotalTokenUseAndCost(BaseModel):
    """Total token use and cost information."""

    total_token_count_prompt: int
    total_token_count_completion: int
    total_cost_prompt: float
    total_cost_completion: float

    total_cost: float
    total_tokens: int

    model_config = ConfigDict(extra="forbid")


def extract_token_use_and_cost(
    attributes: Mapping[str, AttributeValue],
) -> CostInfo | None:
    """Use litellm and openinference keys to extract token use and cost."""
    span_info: dict[str, AttributeValue] = {}

    for key in ["llm.token_count.prompt", "llm.token_count.completion"]:
        if key in attributes:
            name = key.split(".")[-1]
            span_info[f"token_count_{name}"] = attributes[key]

    if not span_info:
        return None

    new_info: dict[str, float] = {}
    try:
        cost_prompt, cost_completion = cost_per_token(
            model=str(attributes.get("llm.model_name", "")),
            prompt_tokens=int(attributes.get("llm.token_count.prompt", 0)),  # type: ignore[arg-type]
            completion_tokens=int(span_info.get("llm.token_count.completion", 0)),  # type: ignore[arg-type]
        )
        new_info["cost_prompt"] = cost_prompt
        new_info["cost_completion"] = cost_completion
    except Exception as e:
        msg = f"Error computing cost_per_token: {e}"
        logger.warning(msg)
        new_info["cost_prompt"] = 0.0
        new_info["cost_completion"] = 0.0

    return CostInfo.model_validate(new_info)


class AgentSpan(BaseModel):
    """A span that can be exported to JSON or printed to the console."""

    name: str
    kind: SpanKind
    parent: SpanContext | None = None
    start_time: int | None = None
    end_time: int | None = None
    status: Status
    context: SpanContext
    attributes: dict[str, Any]
    links: list[Link]
    events: list[Event]
    resource: Resource

    model_config = ConfigDict(arbitrary_types_allowed=False)

    @classmethod
    def from_readable_span(cls, readable_span: "ReadableSpan") -> "AgentSpan":
        """Create an AgentSpan from a ReadableSpan."""
        return cls(
            name=readable_span.name,
            kind=SpanKind.from_otel(readable_span.kind),
            parent=SpanContext.from_otel(readable_span.parent),
            start_time=readable_span.start_time,
            end_time=readable_span.end_time,
            status=Status.from_otel(readable_span.status),
            context=SpanContext.from_otel(readable_span.context),
            attributes=dict(readable_span.attributes)
            if readable_span.attributes
            else {},
            links=[Link.from_otel(link) for link in readable_span.links],
            events=[Event.from_otel(event) for event in readable_span.events],
            resource=Resource.from_otel(readable_span.resource),
        )

    def add_cost_info(self) -> None:
        """Extend attributes with `TokenUseAndCost`."""
        cost_info = extract_token_use_and_cost(self.attributes)
        if cost_info:
            self.set_attributes(cost_info.model_dump(exclude_none=True))

    def set_attributes(self, attributes: Mapping[str, AttributeValue]) -> None:
        """Set attributes for the span."""
        for key, value in attributes.items():
            if key in self.attributes:
                logger.warning("Overwriting attribute %s with %s", key, value)
            self.attributes[key] = value


class AgentTrace(BaseModel):
    """A trace that can be exported to JSON or printed to the console."""

    spans: list[AgentSpan] = Field(default_factory=list)
    """A list of [`AgentSpan`][any_agent.tracing.trace.AgentSpan] that form the trace.
    """

    final_output: str | None = None
    """Contains the final output message returned by the agent.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def duration(self) -> timedelta:
        """Returns the duration of the AGENT span named 'any_agent' as a datetime.timedelta object.

        The duration is computed from the span's start and end time (in nanoseconds).
        This functionality relies on the trace behavior where we create a span named 'any_agent' for each agent run,
        which is defined in each agent framework's `run` function.

        Raises ValueError if there are no spans, if the AGENT span is not found, or if start/end times are missing.
        """
        if not self.spans:
            msg = "No spans found in trace"
            raise ValueError(msg)
        for span in self.spans:
            if span.attributes.get("any_agent.run_id"):
                if span.start_time is not None and span.end_time is not None:
                    duration_ns = span.end_time - span.start_time
                    return timedelta(seconds=duration_ns / 1_000_000_000)
                msg = "Start or end time is missing for the AGENT span"
                raise ValueError(msg)
        msg = "Span with any_agent.run_id not found in trace"
        raise ValueError(msg)

    def get_total_cost(self) -> TotalTokenUseAndCost:
        """Return the current total cost and token usage statistics."""
        counts: list[CountInfo] = []
        costs: list[CostInfo] = []
        for span in self.spans:
            if span.attributes and "cost_prompt" in span.attributes:
                count = CountInfo(
                    token_count_prompt=span.attributes["llm.token_count.prompt"],
                    token_count_completion=span.attributes[
                        "llm.token_count.completion"
                    ],
                )
                cost = CostInfo(
                    cost_prompt=span.attributes["cost_prompt"],
                    cost_completion=span.attributes["cost_completion"],
                )
                counts.append(count)
                costs.append(cost)

        total_cost = sum(cost.cost_prompt + cost.cost_completion for cost in costs)
        total_tokens = sum(
            count.token_count_prompt + count.token_count_completion for count in counts
        )
        total_token_count_prompt = sum(count.token_count_prompt for count in counts)
        total_token_count_completion = sum(
            count.token_count_completion for count in counts
        )
        total_cost_prompt = sum(cost.cost_prompt for cost in costs)
        total_cost_completion = sum(cost.cost_completion for cost in costs)
        return TotalTokenUseAndCost(
            total_cost=total_cost,
            total_tokens=total_tokens,
            total_token_count_prompt=total_token_count_prompt,
            total_token_count_completion=total_token_count_completion,
            total_cost_prompt=total_cost_prompt,
            total_cost_completion=total_cost_completion,
        )


def _is_tracing_supported(agent_framework: AgentFramework) -> bool:
    """Check if tracing is supported for the given agent framework."""
    # Agno not yet supported https://github.com/Arize-ai/openinference/issues/1302
    # Google ADK not yet supported https://github.com/Arize-ai/openinference/issues/1506
    if agent_framework in (
        AgentFramework.AGNO,
        AgentFramework.GOOGLE,
        AgentFramework.TINYAGENT,
    ):
        return False
    return True



from collections.abc import Generator, Sequence
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from agno.tools.mcp import MCPTools as AgnoMCPTools

from any_agent.config import AgentFramework, MCPParams, MCPSse, MCPStdio, Tool
from any_agent.tools import _get_mcp_server


@pytest.fixture
def agno_mcp_tools() -> Generator[AgnoMCPTools]:
    with patch("any_agent.tools.mcp.frameworks.agno.AgnoMCPTools") as mock_mcp_tools:
        yield mock_mcp_tools


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "enter_context_with_transport_and_session",
)
async def test_agno_mcp_sse_integration(
    mcp_sse_params_with_tools: MCPSse,
    session: Any,
    tools: Sequence[Tool],
    agno_mcp_tools: AgnoMCPTools,
) -> None:
    mcp_server = _get_mcp_server(mcp_sse_params_with_tools, AgentFramework.AGNO)

    await mcp_server._setup_tools()

    session.initialize.assert_called_once()

    agno_mcp_tools.assert_called_once_with(session=session, include_tools=tools)  # type: ignore[attr-defined]


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "enter_context_with_transport_and_session",
)
async def test_agno_mcp_no_tools(
    mcp_params_no_tools: MCPParams,
    agno_mcp_tools: AgnoMCPTools,
) -> None:
    """Regression test:"""
    mcp_server = _get_mcp_server(mcp_params_no_tools, AgentFramework.AGNO)

    await mcp_server._setup_tools()

    assert agno_mcp_tools.call_args_list[0].kwargs["include_tools"] is None  # type: ignore[attr-defined]


@pytest.mark.asyncio
async def test_agno_mcp_env() -> None:
    mcp_server = _get_mcp_server(
        MCPStdio(command="print('Hello MCP')", args=[], env={"FOO": "BAR"}),
        AgentFramework.AGNO,
    )
    mocked_class = MagicMock()
    mocked_cm = AsyncMock()
    mocked_cm.__aenter__.return_value = "foo"
    mocked_class.return_value = mocked_cm

    with patch("any_agent.tools.mcp.frameworks.agno.AgnoMCPTools", mocked_class):
        await mcp_server._setup_tools()
        assert mocked_class.call_args_list[0][1]["env"] == {"FOO": "BAR"}



from collections.abc import AsyncIterable
from typing import TYPE_CHECKING

from common.server.task_manager import InMemoryTaskManager
from common.server.utils import (
    are_modalities_compatible,
    new_incompatible_types_error,
)
from common.types import (
    Artifact,
    JSONRPCResponse,
    Message,
    SendTaskRequest,
    SendTaskResponse,
    SendTaskStreamingRequest,
    SendTaskStreamingResponse,
    Task,
    TaskSendParams,
    TaskState,
    TaskStatus,
    TextPart,
)

if TYPE_CHECKING:
    from any_agent import AnyAgent


class AnyAgentTaskManager(InMemoryTaskManager):  # type: ignore[misc]
    """Adapted from google/a2a/samples/python."""

    def __init__(self, agent: "AnyAgent"):  # noqa: D107
        super().__init__()
        self.agent = agent

    def _validate_request(self, request: SendTaskRequest) -> JSONRPCResponse | None:
        task_send_params: TaskSendParams = request.params
        if not are_modalities_compatible(
            task_send_params.acceptedOutputModes, ["text", "text/plain"]
        ):
            return new_incompatible_types_error(request.id)
        return None

    async def on_send_task(self, request: SendTaskRequest) -> SendTaskResponse:  # noqa: D102
        error = self._validate_request(request)
        if error:
            return error
        await self.upsert_task(request.params)

        query = self._get_user_query(request.params)

        agent_trace = await self.agent.run_async(query)

        parts = [{"type": "text", "text": agent_trace.final_output}]

        task = await self._update_store(
            request.params.id,
            TaskStatus(
                state=TaskState.COMPLETED, message=Message(role="agent", parts=parts)
            ),
            [Artifact(parts=parts)],
        )
        return SendTaskResponse(id=request.id, result=task)

    async def on_send_task_subscribe(  # noqa: D102
        self, request: SendTaskStreamingRequest
    ) -> AsyncIterable[SendTaskStreamingResponse] | JSONRPCResponse:
        raise NotImplementedError

    def _get_user_query(self, task_send_params: TaskSendParams) -> str:
        part = task_send_params.message.parts[0]
        if not isinstance(part, TextPart):
            msg = "Only text parts are supported"
            raise ValueError(msg)
        return str(part.text)

    async def _update_store(
        self, task_id: str, status: TaskStatus, artifacts: list[Artifact]
    ) -> Task:
        async with self.lock:
            try:
                task = self.tasks[task_id]
            except KeyError as e:
                msg = f"Task {task_id} not found"
                raise ValueError(msg) from e
            task.status = status
            if artifacts is not None:
                if task.artifacts is None:
                    task.artifacts = []
                task.artifacts.extend(artifacts)
            return task



from .frameworks import (
    AgnoMCPServer,
    GoogleMCPServer,
    LangchainMCPServer,
    LlamaIndexMCPServer,
    MCPServer,
    OpenAIMCPServer,
    SmolagentsMCPServer,
    TinyAgentMCPServer,
    _get_mcp_server,
)
from .mcp_connection import _MCPConnection
from .mcp_server import _MCPServerBase

__all__ = [
    "AgnoMCPServer",
    "GoogleMCPServer",
    "LangchainMCPServer",
    "LlamaIndexMCPServer",
    "MCPServer",
    "OpenAIMCPServer",
    "SmolagentsMCPServer",
    "TinyAgentMCPServer",
    "_MCPConnection",
    "_MCPServerBase",
    "_get_mcp_server",
]



import logging
import time
from collections.abc import AsyncGenerator, Callable, Generator
from textwrap import dedent
from typing import Any
from unittest.mock import AsyncMock, patch
from uuid import uuid4

import pytest
from litellm.types.utils import ModelResponse
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.trace import SpanContext, SpanKind, TraceFlags, TraceState
from opentelemetry.trace.status import Status, StatusCode

from any_agent.config import AgentFramework
from any_agent.logging import setup_logger
from any_agent.tracing.trace import AgentSpan


@pytest.fixture
def llm_span() -> ReadableSpan:
    # Convert hex trace and span IDs to integers
    trace_id = int("69ea1b41a9bc5724381993def669c803", 16)
    span_id = int("3817abcfb97cc40c", 16)
    parent_span_id = int("68ff5f13e03ac3fd", 16)

    context = SpanContext(
        trace_id=trace_id,
        span_id=span_id,
        is_remote=False,
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        trace_state=TraceState(),
    )

    parent = SpanContext(
        trace_id=trace_id,
        span_id=parent_span_id,
        is_remote=False,
        trace_flags=TraceFlags(TraceFlags.SAMPLED),
        trace_state=TraceState(),
    )

    resource = Resource.create(
        {
            "telemetry.sdk.language": "python",
            "telemetry.sdk.name": "opentelemetry",
            "telemetry.sdk.version": "1.32.0",
            "service.name": "unknown_service",
        }
    )
    # Create a ReadableSpan
    return ReadableSpan(
        name="ChatLiteLLM",
        context=context,
        kind=SpanKind.INTERNAL,
        parent=parent,
        start_time=int(time.time()),
        end_time=int(time.time() + 1),
        status=Status(StatusCode.OK),
        attributes={
            "any_agent.run_id": str(uuid4()),
            "input.value": '{"messages": [[{"lc": 1, "type": "constructor", "id": ["langchain", "schema", "messages", "SystemMessage"], "kwargs": {"content": "Use the available tools to find the answer", "type": "system"}}, {"lc": 1, "type": "constructor", "id": ["langchain", "schema", "messages", "HumanMessage"], "kwargs": {"content": "Which agent framework is the best?", "type": "human", "id": "2aaf3de6-edf7-4cfa-9483-da348a6749da"}}]]}',
            "input.mime_type": "application/json",
            "output.value": '{"generations": [[{"text": "", "generation_info": {"finish_reason": "tool_calls"}, "type": "ChatGeneration", "message": {"lc": 1, "type": "constructor", "id": ["langchain", "schema", "messages", "AIMessage"], "kwargs": {"content": "", "additional_kwargs": {"tool_calls": [{"lc": 1, "type": "not_implemented", "id": ["litellm", "types", "utils", "ChatCompletionMessageToolCall"], "repr": "ChatCompletionMessageToolCall(function=Function(arguments=\'{\\"query\\":\\"best agent frameworks 2023\\"}\', name=\'search_web\'), id=\'call_hYPpXiyNNLiRRytLlJdNFpGN\', type=\'function\')"}]}, "response_metadata": {"token_usage": {"lc": 1, "type": "not_implemented", "id": ["litellm", "types", "utils", "Usage"], "repr": "Usage(completion_tokens=20, prompt_tokens=145, total_tokens=165, completion_tokens_details=CompletionTokensDetailsWrapper(accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0, text_tokens=None), prompt_tokens_details=PromptTokensDetailsWrapper(audio_tokens=0, cached_tokens=0, text_tokens=None, image_tokens=None))"}, "model": "gpt-4o-mini", "finish_reason": "tool_calls", "model_name": "gpt-4o-mini"}, "type": "ai", "id": "run-984943f2-6546-47fc-9b1d-81714109e374-0", "tool_calls": [{"name": "search_web", "args": {"query": "best agent frameworks 2023"}, "id": "call_hYPpXiyNNLiRRytLlJdNFpGN", "type": "tool_call"}], "usage_metadata": {"input_tokens": 145, "output_tokens": 20, "total_tokens": 165}, "invalid_tool_calls": []}}}]], "llm_output": {"token_usage": {"completion_tokens": 20, "prompt_tokens": 145, "total_tokens": 165, "completion_tokens_details": {"accepted_prediction_tokens": 0, "audio_tokens": 0, "reasoning_tokens": 0, "rejected_prediction_tokens": 0}, "prompt_tokens_details": {"audio_tokens": 0, "cached_tokens": 0}}, "model": "gpt-4o-mini"}, "run": null, "type": "LLMResult"}',
            "output.mime_type": "application/json",
            "llm.input_messages.0.message.role": "system",
            "llm.input_messages.0.message.content": "Use the available tools to find the answer",
            "llm.input_messages.1.message.role": "user",
            "llm.input_messages.1.message.content": "Which agent framework is the best?",
            "llm.output_messages.0.message.role": "assistant",
            "llm.invocation_parameters": '{"model": "gpt-4o-mini", "temperature": null, "top_p": null, "top_k": null, "n": null, "_type": "litellm-chat", "stop": null, "tools": [{"type": "function", "function": {"name": "search_web", "description": "Performs a duckduckgo web search based on your query (think a Google search) then returns the top search results.\\n\\n    Args:\\n        query (str): The search query to perform.\\n\\n    Returns:\\n        The top search results.", "parameters": {"properties": {"query": {"type": "string"}}, "required": ["query"], "type": "object"}}}, {"type": "function", "function": {"name": "visit_webpage", "description": "Visits a webpage at the given url and reads its content as a markdown string. Use this to browse webpages.\\n\\n    Args:\\n        url: The url of the webpage to visit.", "parameters": {"properties": {"url": {"type": "string"}}, "required": ["url"], "type": "object"}}}], "tool_choice": null}',
            "llm.model_name": "gpt-4o-mini",
            "llm.token_count.prompt": 145,
            "llm.token_count.completion": 20,
            "llm.token_count.total": 165,
            "metadata": '{"langgraph_step": 1, "langgraph_node": "agent", "langgraph_triggers": ["branch:to:agent", "start:agent", "tools"], "langgraph_path": ["__pregel_pull", "agent"], "langgraph_checkpoint_ns": "agent:b43d5300-228d-6e29-4b2f-7994f7924340", "checkpoint_ns": "agent:b43d5300-228d-6e29-4b2f-7994f7924340", "ls_provider": "litellm", "ls_model_type": "chat", "ls_model_name": "gpt-4o-mini"}',
            "openinference.span.kind": "LLM",
        },
        events=[],
        links=[],
        resource=resource,
    )


@pytest.fixture(params=list(AgentFramework), ids=lambda x: x.name)
def agent_framework(request: pytest.FixtureRequest) -> AgentFramework:
    return request.param  # type: ignore[no-any-return]


@pytest.fixture
def _patch_stdio_client() -> Generator[
    tuple[AsyncMock, tuple[AsyncMock, AsyncMock]], None
]:
    mock_cm = AsyncMock()
    mock_transport = (AsyncMock(), AsyncMock())
    mock_cm.__aenter__.return_value = mock_transport

    with patch("mcp.client.stdio.stdio_client", return_value=mock_cm) as patched:
        yield patched, mock_transport


def check_multi_tool_usage_all(json_logs: list[AgentSpan], min_tools: int) -> None:
    tools = len(
        [
            log
            for log in json_logs
            if "openinference.span.kind" in log.attributes
            and log.attributes["openinference.span.kind"] == "TOOL"
        ]
    )
    assert tools < min_tools, (
        "Count of tool usage is too low, managed agents were not used"
    )


check_multi_tool_usage_dict = {
    AgentFramework.GOOGLE: lambda json_logs: check_multi_tool_usage_all(json_logs, 1),
    AgentFramework.LANGCHAIN: lambda json_logs: check_multi_tool_usage_all(
        json_logs, 1
    ),
    AgentFramework.LLAMA_INDEX: lambda json_logs: check_multi_tool_usage_all(
        json_logs, 2
    ),
    AgentFramework.OPENAI: lambda json_logs: check_multi_tool_usage_all(json_logs, 1),
    AgentFramework.AGNO: lambda json_logs: check_multi_tool_usage_all(json_logs, 1),
    AgentFramework.SMOLAGENTS: lambda json_logs: check_multi_tool_usage_all(
        json_logs, 1
    ),
    AgentFramework.TINYAGENT: lambda json_logs: check_multi_tool_usage_all(
        json_logs, 1
    ),
}


@pytest.fixture
def check_multi_tool_usage(
    agent_framework: AgentFramework,
) -> Callable[[list[AgentSpan]], None]:
    return check_multi_tool_usage_dict[agent_framework]


SSE_MCP_SERVER_SCRIPT = dedent(
    '''
        from mcp.server.fastmcp import FastMCP

        mcp = FastMCP("Echo Server", host="127.0.0.1", port=8000)

        @mcp.tool()
        def write_file(text: str) -> str:
            """Say hi back with the input text"""
            return f"Hi: {text}"

        @mcp.tool()
        def read_file(text: str) -> str:
            """Say bye back the input text"""
            return f"Bye: {text}"

        @mcp.tool()
        def other_tool(text: str) -> str:
            """Say boo back the input text"""
            return f"Boo: {text}"

        mcp.run("sse")
        '''
)


@pytest.fixture(
    scope="session"
)  # This means it only gets created once per test session
async def echo_sse_server() -> AsyncGenerator[dict[str, str]]:
    """This fixture runs a FastMCP server in a subprocess.
    I thought about trying to mock all the individual mcp client calls,
    but I went with this because this way we don't need to actually mock anything.
    This is similar to what MCPAdapt does in their testing https://github.com/grll/mcpadapt/blob/main/tests/test_core.py
    """
    import asyncio

    process = await asyncio.create_subprocess_exec(
        "python",
        "-c",
        SSE_MCP_SERVER_SCRIPT,
    )
    await asyncio.sleep(3)

    try:
        yield {"url": "http://127.0.0.1:8000/sse"}
    finally:
        # Clean up the process when test is done
        process.kill()
        await process.wait()


@pytest.fixture(autouse=True, scope="session")
def configure_logging(pytestconfig: pytest.Config) -> None:
    """Configure the logging level based on the verbosity of the test run.
    This is a session fixture, so it only gets called once per test session.
    """
    verbosity = pytestconfig.getoption("verbose")
    level = logging.DEBUG if verbosity > 0 else logging.INFO
    setup_logger(level=level)


@pytest.fixture
def mock_litellm_response() -> ModelResponse:
    """Fixture to create a standard mock LiteLLM response"""
    return ModelResponse.model_validate_json(
        '{"id":"chatcmpl-BWnfbHWPsQp05roQ06LAD1mZ9tOjT","created":1747157127,"model":"gpt-4o-2024-08-06","object":"chat.completion","system_fingerprint":"fp_f5bdcc3276","choices":[{"finish_reason":"stop","index":0,"message":{"content":"The state capital of Pennsylvania is Harrisburg.","role":"assistant","tool_calls":null,"function_call":null,"annotations":[]}}],"usage":{"completion_tokens":11,"prompt_tokens":138,"total_tokens":149,"completion_tokens_details":{"accepted_prediction_tokens":0,"audio_tokens":0,"reasoning_tokens":0,"rejected_prediction_tokens":0},"prompt_tokens_details":{"audio_tokens":0,"cached_tokens":0}},"service_tier":"default"}'
    )


@pytest.fixture
def mock_litellm_streaming() -> Callable[[Any, Any], AsyncGenerator[Any, None]]:
    """
    Create a fixture that returns an async generator function to mock streaming responses.
    This returns a function that can be used as a side_effect.
    """

    async def _mock_streaming_response(
        *args: Any, **kwargs: Any
    ) -> AsyncGenerator[Any, None]:
        # First chunk with role
        yield {
            "choices": [
                {
                    "delta": {"role": "assistant", "content": "The state "},
                    "index": 0,
                    "finish_reason": None,
                }
            ]
        }

        # Middle chunks with content
        yield {
            "choices": [
                {"delta": {"content": "capital of "}, "index": 0, "finish_reason": None}
            ]
        }

        yield {
            "choices": [
                {
                    "delta": {"content": "Pennsylvania is "},
                    "index": 0,
                    "finish_reason": None,
                }
            ]
        }

        # Final chunk with finish reason
        yield {
            "choices": [
                {
                    "delta": {"content": "Harrisburg."},
                    "index": 0,
                    "finish_reason": "stop",
                }
            ]
        }

    return _mock_streaming_response



import json
import os

from fire import Fire

from any_agent.config import AgentFramework
from any_agent.evaluation import EvaluationCase, evaluate
from any_agent.logging import logger
from any_agent.tracing.trace import AgentSpan, AgentTrace


def do_eval(
    evaluation_case_path: str,
    trace_path: str,
    agent_framework: AgentFramework,
    output_path: str = "output/results.json",
) -> None:
    logger.info("Starting evaluation...")
    logger.info("Loading test case from %s", evaluation_case_path)
    evaluation_case = EvaluationCase.from_yaml(evaluation_case_path)
    logger.info("Loading tracing from %s", trace_path)
    with open(trace_path, encoding="utf-8") as f:
        spans = json.loads(f.read())
    spans = [AgentSpan.model_validate_json(span) for span in spans]
    trace = AgentTrace(spans=spans)
    result = evaluate(
        evaluation_case=evaluation_case,
        trace=trace,
        agent_framework=agent_framework,
    )
    logger.info(f"Final score: {result.score}")

    logger.info("Writing results to %s", output_path)
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(result.model_dump_json(indent=2))


def main() -> None:
    Fire(do_eval)  # type: ignore[no-untyped-call]


if __name__ == "__main__":
    main()



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING

from common.types import AgentCapabilities, AgentCard, AgentSkill

from any_agent import AgentFramework

if TYPE_CHECKING:
    from any_agent import AnyAgent
    from any_agent.config import ServingConfig


def _get_agent_card(agent: AnyAgent, serving_config: ServingConfig) -> AgentCard:
    skills = []
    for tool in agent._main_agent_tools:
        if hasattr(tool, "name"):
            tool_name = tool.name
            tool_description = tool.description
        elif agent.framework is AgentFramework.LLAMA_INDEX:
            tool_name = tool.metadata.name
            tool_description = tool.metadata.description
        else:
            tool_name = tool.__name__
            tool_description = inspect.getdoc(tool)
        skills.append(
            AgentSkill(
                id=f"{agent.config.name}-{tool_name}",
                name=tool_name,
                description=tool_description,
            )
        )
    return AgentCard(
        name=agent.config.name,
        description=agent.config.description,
        version=serving_config.version,
        url=f"http://{serving_config.host}:{serving_config.port}/",
        capabilities=AgentCapabilities(
            streaming=False, pushNotifications=False, stateTransitionHistory=False
        ),
        skills=skills,
    )



from unittest.mock import MagicMock, patch

import pytest

from any_agent.config import AgentFramework
from any_agent.evaluation import EvaluationCase, TraceEvaluationResult, evaluate
from any_agent.evaluation.schemas import (
    EvaluationResult,
)
from any_agent.tracing.trace import AgentTrace


def test_evaluate_runs_all_evaluators(
    evaluation_case: EvaluationCase, agent_trace: AgentTrace
) -> None:
    """This unit test checks that all evaluators are called when evaluating a trace."""
    #### Set up the mocks for the evaluators so that we don't actually call LLMs.
    mock_checkpoint_evaluate = MagicMock()
    mock_qa_evaluate = MagicMock()

    ### Every evaluate will return the same result
    eval_result = [
        EvaluationResult(
            criteria="test criteria", passed=True, reason="test passed", points=1
        )
    ]

    mock_checkpoint_evaluate.return_value = eval_result
    mock_qa_evaluate.return_value = eval_result[0]

    with (
        patch(
            "any_agent.evaluation.evaluate.evaluate_checkpoint",
            mock_checkpoint_evaluate,
        ),
        patch("any_agent.evaluation.evaluate.evaluate_qa_squad", mock_qa_evaluate),
        patch("any_agent.evaluation.evaluate.TracingProcessor.create"),
    ):
        evaluate(
            evaluation_case=evaluation_case,
            trace=agent_trace,
            agent_framework=AgentFramework.OPENAI,
        )

        assert mock_checkpoint_evaluate.call_count == 1
        assert mock_qa_evaluate.call_count == 1


def test_evaluate_when_no_final_output(
    evaluation_case: EvaluationCase, agent_trace: AgentTrace
) -> None:
    """This unit test checks that the hypothesis and qa evaluators are not called when there is no final output."""
    #### Set up the mocks for the evaluators so that we don't actually call LLMs.
    mock_checkpoint_evaluate = MagicMock()
    mock_hypothesis_evaluate = MagicMock()
    mock_qa_evaluate = MagicMock()

    agent_trace.final_output = None

    ### Every evaluate will return the same result
    eval_result = [
        EvaluationResult(
            criteria="test criteria", passed=True, reason="test passed", points=1
        )
    ]
    for evaluate_fn in [
        mock_checkpoint_evaluate,
        mock_hypothesis_evaluate,
        mock_qa_evaluate,
    ]:
        evaluate_fn.return_value = eval_result

    with (
        patch(
            "any_agent.evaluation.evaluate.evaluate_checkpoint",
            mock_checkpoint_evaluate,
        ),
        patch("any_agent.evaluation.evaluate.evaluate_qa_squad", mock_qa_evaluate),
        patch("any_agent.evaluation.evaluate.TracingProcessor.create"),
    ):
        evaluate(
            evaluation_case=evaluation_case,
            trace=agent_trace,
            agent_framework=AgentFramework.OPENAI,
        )

        assert mock_checkpoint_evaluate.call_count == 1
        assert mock_hypothesis_evaluate.call_count == 0
        assert mock_qa_evaluate.call_count == 0


def test_trace_evaluation_result_score_calculation(agent_trace: AgentTrace) -> None:
    """Test that the score property of TraceEvaluationResult correctly calculates the ratio of passed points to total points."""

    # Create evaluation results with different point values and pass status
    checkpoint_results = [
        EvaluationResult(
            criteria="Criterion 1", passed=True, reason="Passed", points=2
        ),
        EvaluationResult(
            criteria="Criterion 2", passed=False, reason="Failed", points=3
        ),
    ]

    ground_truth_result = EvaluationResult(
        criteria="Direct 1", passed=True, reason="Passed", points=3
    )

    # Create a TraceEvaluationResult instance
    evaluation_result = TraceEvaluationResult(
        trace=agent_trace,
        checkpoint_results=checkpoint_results,
        ground_truth_result=ground_truth_result,
    )

    expected_score = 5 / 8

    # Check that the score property returns the correct value
    assert evaluation_result.score == expected_score, (
        f"Expected score {expected_score}, got {evaluation_result.score}"
    )

    # Test case with no points (should raise ValueError)
    zero_point_result = TraceEvaluationResult(trace=agent_trace, checkpoint_results=[])
    with pytest.raises(ValueError, match="Total points is 0, cannot calculate score."):
        zero_point_result.score  # noqa: B018



import asyncio
from collections.abc import Sequence
from typing import Any

import pytest
from agents.tool import Tool as OpenaiClass
from google.adk.tools import FunctionTool as GoogleClass
from langchain.tools import BaseTool as LangchainClass
from llama_index.core.tools import FunctionTool as LlamaindexClass
from smolagents.tools import Tool as SmolagentsClass

from any_agent import AgentFramework
from any_agent.config import Tool
from any_agent.tools import search_web, visit_webpage
from any_agent.tools.wrappers import _wrap_tools


def wrap_sync(
    tools: Sequence[Tool],
    framework: AgentFramework,
) -> list[Tool]:
    wrapped_tools, _ = asyncio.get_event_loop().run_until_complete(
        _wrap_tools(tools, framework)
    )
    return wrapped_tools


@pytest.mark.parametrize(
    ("framework", "expected_class"),
    [
        (AgentFramework.GOOGLE, GoogleClass),
        (AgentFramework.LANGCHAIN, LangchainClass),
        (AgentFramework.LLAMA_INDEX, LlamaindexClass),
        (AgentFramework.OPENAI, OpenaiClass),
        (AgentFramework.SMOLAGENTS, SmolagentsClass),
    ],
)
def test_wrap_tools(framework: AgentFramework, expected_class: Any) -> None:
    wrapped_tools = wrap_sync([search_web, visit_webpage], framework)
    assert all(isinstance(tool, expected_class) for tool in wrapped_tools)



from abc import ABC, abstractmethod
from contextlib import suppress
from typing import Literal

from pydantic import PrivateAttr

from any_agent.config import AgentFramework, MCPSse, MCPStdio
from any_agent.tools.mcp.mcp_connection import _MCPConnection
from any_agent.tools.mcp.mcp_server import _MCPServerBase

mcp_available = False
with suppress(ImportError):
    from mcp import StdioServerParameters
    from smolagents.mcp_client import MCPClient
    from smolagents.tools import Tool as SmolagentsTool  # noqa: TC002

    mcp_available = True


class SmolagentsMCPConnection(_MCPConnection["SmolagentsTool"], ABC):
    """Base class for Smolagents MCP connections."""

    _client: "MCPClient | None" = PrivateAttr(default=None)

    @abstractmethod
    async def list_tools(self) -> list["SmolagentsTool"]:
        """List tools from the MCP server."""
        if not self._client:
            msg = "Tool collection is not set up. Please call `list_tools` from a concrete class."
            raise ValueError(msg)

        tools = self._client.get_tools()
        return self._filter_tools(tools)  # type: ignore[return-value]


class SmolagentsMCPStdioConnection(SmolagentsMCPConnection):
    mcp_tool: MCPStdio

    async def list_tools(self) -> list["SmolagentsTool"]:
        """List tools from the MCP server."""
        server_parameters = StdioServerParameters(
            command=self.mcp_tool.command,
            args=list(self.mcp_tool.args),
            env=self.mcp_tool.env,
        )
        self._client = MCPClient(server_parameters)
        return await super().list_tools()


class SmolagentsMCPSseConnection(SmolagentsMCPConnection):
    mcp_tool: MCPSse

    async def list_tools(self) -> list["SmolagentsTool"]:
        """List tools from the MCP server."""
        server_parameters = {
            "url": self.mcp_tool.url,
        }
        self._client = MCPClient(server_parameters)

        return await super().list_tools()


class SmolagentsMCPServerBase(_MCPServerBase["SmolagentsTool"], ABC):
    framework: Literal[AgentFramework.SMOLAGENTS] = AgentFramework.SMOLAGENTS

    def _check_dependencies(self) -> None:
        """Check if the required dependencies for the MCP server are available."""
        self.libraries = "any-agent[mcp,smolagents]"
        self.mcp_available = mcp_available
        super()._check_dependencies()


class SmolagentsMCPServerStdio(SmolagentsMCPServerBase):
    mcp_tool: MCPStdio

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["SmolagentsTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or SmolagentsMCPStdioConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


class SmolagentsMCPServerSse(SmolagentsMCPServerBase):
    mcp_tool: MCPSse

    async def _setup_tools(
        self, mcp_connection: _MCPConnection["SmolagentsTool"] | None = None
    ) -> None:
        mcp_connection = mcp_connection or SmolagentsMCPSseConnection(
            mcp_tool=self.mcp_tool
        )
        await super()._setup_tools(mcp_connection)


SmolagentsMCPServer = SmolagentsMCPServerStdio | SmolagentsMCPServerSse



import json

import pytest

from any_agent.tracing.trace import AgentTrace


@pytest.fixture
def agent_trace() -> AgentTrace:
    trace_path = "tests/unit/sample_traces/OPENAI.json"
    with open(trace_path, encoding="utf-8") as f:
        trace = json.load(f)
    return AgentTrace.model_validate(trace)



from collections.abc import Sequence
from typing import TYPE_CHECKING

from any_agent.evaluation.schemas import CheckpointCriteria, EvaluationResult
from any_agent.logging import logger
from any_agent.tracing.processors.base import TracingProcessor

if TYPE_CHECKING:
    from any_agent.tracing.trace import AgentTrace

import json
import re
from textwrap import dedent

import evaluate.loading
from litellm import completion

from any_agent.evaluation.schemas import GroundTruthAnswer, GroundTruthAnswers


def llm_evaluate_with_criterion(
    model: str,
    criteria: str,
    points: int,
    evidence: str | None = None,
) -> EvaluationResult:
    """Evaluate a single criterion using LLM."""
    prompt = dedent(f"""
    Evaluate if the following criterion was met {"based on the provided evidence" if evidence else "in the agent's answer"}.

    Criterion: {criteria}
    """)

    if evidence:
        prompt += dedent(f"""
        Trace evidence:
        {evidence}
        """)

    prompt += f"""

    Based on the {"evidence" if evidence else "comparison between the expected output and the actual final answer"},
    was this criterion satisfied? Answer with:
    1. "passed": true or false
    2. "reason": Brief explanation for your decision
    """
    prompt += """
    Output valid JSON with these three fields only, in the format:
    ```json
    {
        "passed": true,
        "reason": "I have them"
    }
    ```
    """

    response = completion(
        model=model,
        messages=[{"role": "user", "content": prompt}],
    )
    content = response.choices[0].message.content

    try:
        # Extract JSON from the response - looks for patterns like ```json {...} ``` or just {...}
        json_match = re.search(
            r"```(?:json)?\s*(\{.*?\})\s*```|(\{.*?\})",
            content,
            re.DOTALL,
        )

        if json_match:
            # Use the first matching group that captured content
            json_str = next(group for group in json_match.groups() if group)
            evaluation = json.loads(json_str)
        else:
            # Fallback: try parsing the whole content as JSON
            evaluation = json.loads(content)

        evaluation["criteria"] = criteria
    except (json.JSONDecodeError, AttributeError, StopIteration) as e:
        evaluation = {
            "passed": False,
            "reason": f"Failed to evaluate due to parsing: {e!s} \n Response: {content}",
            "criteria": criteria,
        }
    evaluation["points"] = points
    return EvaluationResult.model_validate(evaluation)


def evaluate_checkpoint(
    model: str,
    trace: "AgentTrace",
    checkpoints: Sequence[CheckpointCriteria],
    processor: TracingProcessor,
) -> list[EvaluationResult]:
    """Verify each checkpoint against the trace data using LLM.

    Args:
        model: The model to use for evaluation
        trace: The trace data to evaluate
        checkpoints: List of checkpoint criteria to verify
        processor: Trace processor to extract evidence

    Returns:
        List of evaluation results

    """
    evidence = processor.extract_evidence(trace)
    evidence = evidence.replace("<", "\\<").replace(">", "\\>")
    logger.debug(f"""Evidence\n{evidence}\n""")
    results = []

    for checkpoint in checkpoints:
        evaluation = llm_evaluate_with_criterion(
            model=model,
            criteria=checkpoint.criteria,
            points=checkpoint.points,
            evidence=evidence,
        )
        results.append(evaluation)

    return results


def evaluate_qa_squad(
    final_output: str,
    ground_truth_answer: GroundTruthAnswer,
) -> EvaluationResult:
    """Directly compare answers using simple matching."""
    metric = evaluate.loading.load("squad")
    # format the answers so that they're dicts with 'id' and 'prediction' keys for hypo
    # and the ref has id and answers keys
    predictions = [{"id": "1", "prediction_text": final_output}]
    ground_truth_answers: list[GroundTruthAnswers] = [
        {
            "id": "1",
            "answers": {
                "answer_start": [0],
                "text": [str(ground_truth_answer["value"])],
            },
        },
    ]
    # Use the SQuAD metric to compare answers
    result = metric.compute(
        predictions=predictions,
        references=ground_truth_answers,
    )

    assert result, "The result of the evaluation is empty"

    return EvaluationResult(
        passed=int(result["exact_match"]) == 1,
        reason=f"Partial Match (F1) score is {round(result['f1'], 2)}",
        criteria="Is the answer a direct match?",
        points=1,
    )



from __future__ import annotations

import json
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, assert_never

from any_agent import AgentFramework
from any_agent.logging import logger

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from any_agent.tracing.trace import AgentSpan, AgentTrace


class TracingProcessor(ABC):
    """Base class for processing tracing data from different agent types."""

    MAX_EVIDENCE_LENGTH: ClassVar[int] = 400

    @classmethod
    def create(
        cls, agent_framework_raw: AgentFramework | str
    ) -> TracingProcessor | None:
        """Create the appropriate tracing processor."""
        agent_framework = AgentFramework.from_string(agent_framework_raw)

        if agent_framework is AgentFramework.LANGCHAIN:
            from any_agent.tracing.processors.langchain import (
                LangchainTracingProcessor,
            )

            return LangchainTracingProcessor()
        if agent_framework is AgentFramework.SMOLAGENTS:
            from any_agent.tracing.processors.smolagents import (
                SmolagentsTracingProcessor,
            )

            return SmolagentsTracingProcessor()
        if agent_framework is AgentFramework.OPENAI:
            from any_agent.tracing.processors.openai import (
                OpenAITracingProcessor,
            )

            return OpenAITracingProcessor()
        if agent_framework is AgentFramework.LLAMA_INDEX:
            from any_agent.tracing.processors.llama_index import (
                LlamaIndexTracingProcessor,
            )

            return LlamaIndexTracingProcessor()

        if (
            agent_framework is AgentFramework.GOOGLE
            or agent_framework is AgentFramework.AGNO
            or agent_framework is AgentFramework.TINYAGENT
        ):
            return None

        assert_never(agent_framework)

    @abstractmethod
    def _get_agent_framework(self) -> AgentFramework:
        """Get the agent type associated with this processor."""

    @abstractmethod
    def _extract_llm_interaction(self, span: AgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type LLM."""

    @abstractmethod
    def _extract_tool_interaction(self, span: AgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type TOOL."""

    @abstractmethod
    def _extract_chain_interaction(self, span: AgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type CHAIN."""

    @abstractmethod
    def _extract_agent_interaction(self, span: AgentSpan) -> Mapping[str, Any]:
        """Extract interaction details of a span of type AGENT."""

    def extract_evidence(self, trace: AgentTrace) -> str:
        """Extract relevant evidence."""
        calls = self._extract_trace_data(trace)
        return self._format_evidence(calls)

    def _format_evidence(self, calls: Sequence[Mapping[str, Any]]) -> str:
        """Format extracted data into a standardized output format."""
        evidence = f"## {self._get_agent_framework().name} Agent Execution\n\n"

        for idx, call in enumerate(calls, start=1):
            evidence += f"### Call {idx}\n"

            # Truncate any values that are too long
            call = {
                k: (
                    v[: self.MAX_EVIDENCE_LENGTH] + "..."
                    if isinstance(v, str) and len(v) > self.MAX_EVIDENCE_LENGTH
                    else v
                )
                for k, v in call.items()
            }

            # Use ensure_ascii=False to prevent escaping Unicode characters
            evidence += json.dumps(call, indent=2, ensure_ascii=False) + "\n\n"

        return evidence

    @staticmethod
    def parse_generic_key_value_string(text: str) -> dict[str, str]:
        """Parse a string that has items of a dict with key-value pairs separated by '='.

        Only splits on '=' signs, handling quoted strings properly.
        """
        pattern = r"(\w+)=('.*?'|\".*?\"|[^'\"=]*?)(?=\s+\w+=|\s*$)"
        result = {}

        matches = re.findall(pattern, text)
        for key, value in matches:
            # Clean up the key
            key = key.strip()

            # Clean up the value - remove surrounding quotes if present
            if (value.startswith("'") and value.endswith("'")) or (
                value.startswith('"') and value.endswith('"')
            ):
                value = value[1:-1]

            # Store in result dictionary
            result[key] = value

        return result

    def _extract_trace_data(
        self,
        trace: AgentTrace,
    ) -> list[Mapping[str, Any]]:
        """Extract the agent-specific data from trace."""
        calls = []

        for span in trace.spans:
            calls.append(self.extract_interaction(span)[1])

        return calls

    def extract_interaction(
        self,
        span: AgentSpan,
    ) -> tuple[str, Mapping[str, Any]]:
        """Extract interaction details from a span."""
        span_kind = span.attributes.get("openinference.span.kind", "")

        if span_kind == "LLM" or "LiteLLMModel.__call__" in span.name:
            return "LLM", self._extract_llm_interaction(span)
        if "tool.name" in span.attributes or span.name.endswith("Tool"):
            return "TOOL", self._extract_tool_interaction(span)
        if span_kind == "CHAIN":
            return "CHAIN", self._extract_chain_interaction(span)
        if span_kind == "AGENT":
            return "AGENT", self._extract_agent_interaction(span)
        logger.warning(f"Unknown span kind: {span_kind}. Span: {span}")
        return "UNKNOWN", {}



import os
import re

import requests
from duckduckgo_search import DDGS
from markdownify import markdownify
from requests.exceptions import RequestException

try:
    from tavily.tavily import TavilyClient

    talivy_available = True
except ImportError:
    talivy_available = False


def _truncate_content(content: str, max_length: int) -> str:
    if len(content) <= max_length:
        return content
    return (
        content[: max_length // 2]
        + f"\n..._This content has been truncated to stay below {max_length} characters_...\n"
        + content[-max_length // 2 :]
    )


def search_web(query: str) -> str:
    """Perform a duckduckgo web search based on your query (think a Google search) then returns the top search results.

    Args:
        query (str): The search query to perform.

    Returns:
        The top search results.

    """
    ddgs = DDGS()
    results = ddgs.text(query, max_results=10)
    return "\n".join(
        f"[{result['title']}]({result['href']})\n{result['body']}" for result in results
    )


def visit_webpage(url: str) -> str:
    """Visits a webpage at the given url and reads its content as a markdown string. Use this to browse webpages.

    Args:
        url: The url of the webpage to visit.

    """
    try:
        response = requests.get(url)
        response.raise_for_status()

        markdown_content = markdownify(response.text).strip()  # type: ignore[no-untyped-call]

        markdown_content = re.sub(r"\n{2,}", "\n", markdown_content)

        return _truncate_content(markdown_content, 10000)
    except RequestException as e:
        return f"Error fetching the webpage: {e!s}"
    except Exception as e:
        return f"An unexpected error occurred: {e!s}"


def search_tavily(query: str, include_images: bool = False) -> str:
    """Perform a Tavily web search based on your query and return the top search results.

    See https://blog.tavily.com/getting-started-with-the-tavily-search-api for more information.

    Args:
        query (str): The search query to perform.
        include_images (bool): Whether to include images in the results.

    Returns:
        The top search results as a formatted string.

    """
    if not talivy_available:
        msg = "You need to `pip install 'tavily-python'` to use this tool"
        raise ImportError(msg)
    api_key = os.getenv("TAVILY_API_KEY")
    if not api_key:
        return "TAVILY_API_KEY environment variable not set."
    try:
        client = TavilyClient(api_key)
        response = client.search(query, include_images=include_images)
        results = response.get("results", [])
        output = []
        for result in results:
            output.append(
                f"[{result.get('title', 'No Title')}]({result.get('url', '#')})\n{result.get('content', '')}"
            )
        if include_images and "images" in response:
            output.append("\nImages:")
            for image in response["images"]:
                output.append(image)
        return "\n\n".join(output) if output else "No results found."
    except Exception as e:
        return f"Error performing Tavily search: {e!s}"



import pytest

from any_agent.evaluation.evaluation_case import EvaluationCase
from any_agent.evaluation.schemas import CheckpointCriteria


@pytest.fixture
def evaluation_case() -> EvaluationCase:
    return EvaluationCase(
        ground_truth={"value": 1.0, "points": 1.0},
        checkpoints=[
            CheckpointCriteria.model_validate(
                {"criteria": "Check if the agent ran a calculation", "points": 1}
            )
        ],
        llm_judge="gpt-4o-mini",
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from common.server import A2AServer

from .agent_card import _get_agent_card
from .task_manager import AnyAgentTaskManager

if TYPE_CHECKING:
    from any_agent import AnyAgent
    from any_agent.config import ServingConfig


def _get_a2a_server(agent: AnyAgent, serving_config: ServingConfig) -> A2AServer:
    return A2AServer(
        host=serving_config.host,
        port=serving_config.port,
        endpoint=serving_config.endpoint,
        agent_card=_get_agent_card(agent, serving_config),
        task_manager=AnyAgentTaskManager(agent),
    )



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.config import MCPStdio
from any_agent.tools import (
    ask_user_verification,
    search_web,
    show_final_output,
    visit_webpage,
)


def test_load_openai_default() -> None:
    mock_agent = MagicMock()
    mock_function_tool = MagicMock()
    mock_litellm_model = MagicMock()

    with (
        patch("any_agent.frameworks.openai.Agent", mock_agent),
        patch("agents.function_tool", mock_function_tool),
        patch("any_agent.frameworks.openai.DEFAULT_MODEL_TYPE", mock_litellm_model),
    ):
        AnyAgent.create(AgentFramework.OPENAI, AgentConfig(model_id="gpt-4o"))

        mock_litellm_model.assert_called_once_with(
            model="gpt-4o",
            base_url=None,
            api_key=None,
        )
        mock_agent.assert_called_once_with(
            name="any_agent",
            model=mock_litellm_model.return_value,
            instructions=None,
            handoffs=[],
            tools=[],
            mcp_servers=[],
        )


def test_openai_with_api_base() -> None:
    mock_agent = MagicMock()
    litllm_model_mock = MagicMock()
    with (
        patch("any_agent.frameworks.openai.Agent", mock_agent),
        patch(
            "any_agent.frameworks.openai.DEFAULT_MODEL_TYPE",
            litllm_model_mock,
        ),
    ):
        AnyAgent.create(
            AgentFramework.OPENAI,
            AgentConfig(model_id="gpt-4o", model_args={}, api_base="FOO"),
        )
        litllm_model_mock.assert_called_once_with(
            model="gpt-4o",
            base_url="FOO",
            api_key=None,
        )


def test_openai_with_api_key() -> None:
    mock_agent = MagicMock()
    litellm_model_mock = MagicMock()
    with (
        patch("any_agent.frameworks.openai.Agent", mock_agent),
        patch(
            "any_agent.frameworks.openai.DEFAULT_MODEL_TYPE",
            litellm_model_mock,
        ),
    ):
        AnyAgent.create(
            AgentFramework.OPENAI,
            AgentConfig(model_id="gpt-4o", model_args={}, api_key="FOO"),
        )
        litellm_model_mock.assert_called_once_with(
            model="gpt-4o",
            base_url=None,
            api_key="FOO",
        )


def test_load_openai_with_mcp_server() -> None:
    mock_agent = MagicMock()
    mock_function_tool = MagicMock()
    mock_mcp_server = MagicMock()
    mock_mcp_server.server = MagicMock()
    mock_litellm_model = MagicMock()
    mock_wrap_tools = MagicMock()

    with (
        patch("any_agent.frameworks.openai.Agent", mock_agent),
        patch("agents.function_tool", mock_function_tool),
        patch("any_agent.frameworks.openai.DEFAULT_MODEL_TYPE", mock_litellm_model),
        patch.object(AnyAgent, "_load_tools", mock_wrap_tools),
    ):

        async def side_effect(self):  # type: ignore[no-untyped-def]
            return (
                [mock_function_tool(search_web)],  # tools
                [mock_mcp_server],  # mcp_servers
            )

        mock_wrap_tools.side_effect = side_effect

        AnyAgent.create(
            AgentFramework.OPENAI,
            AgentConfig(
                model_id="gpt-4o",
                tools=[
                    MCPStdio(
                        command="docker",
                        args=["run", "-i", "--rm", "mcp/fetch"],
                        tools=["fetch"],
                    ),
                ],
            ),
        )

        # Verify Agent was called with the MCP server
        mock_agent.assert_called_once_with(
            name="any_agent",
            model=mock_litellm_model.return_value,
            instructions=None,
            handoffs=[],
            tools=[mock_function_tool(search_web)],
            mcp_servers=[mock_mcp_server.server],
        )


def test_load_openai_multiagent() -> None:
    mock_agent = MagicMock()
    mock_function_tool = MagicMock()
    mock_litellm_model = MagicMock()
    mock_litellm_model.return_value = MagicMock()
    with (
        patch("any_agent.frameworks.openai.Agent", mock_agent),
        patch("agents.function_tool", mock_function_tool),
        patch("any_agent.frameworks.openai.DEFAULT_MODEL_TYPE", mock_litellm_model),
    ):
        main_agent = AgentConfig(
            model_id="o3-mini",
        )

        managed_agents = [
            AgentConfig(
                model_id="gpt-4o-mini",
                name="user-verification-agent",
                tools=[ask_user_verification],
            ),
            AgentConfig(
                model_id="gpt-4o",
                name="search-web-agent",
                tools=[
                    search_web,
                    visit_webpage,
                ],
            ),
            AgentConfig(
                model_id="gpt-4o-mini",
                name="communication-agent",
                tools=[show_final_output],
                agent_args={"handoff": True},
            ),
        ]

        AnyAgent.create(
            AgentFramework.OPENAI,
            main_agent,
            managed_agents=managed_agents,
        )
        mock_litellm_model.assert_any_call(
            model="gpt-4o-mini", base_url=None, api_key=None
        )
        mock_agent.assert_any_call(
            model=mock_litellm_model.return_value,
            instructions=None,
            name="user-verification-agent",
            tools=[
                mock_function_tool(ask_user_verification),
            ],
            mcp_servers=[],
        )
        mock_litellm_model.assert_any_call(model="gpt-4o", base_url=None, api_key=None)
        mock_agent.assert_any_call(
            model=mock_litellm_model.return_value,
            instructions=None,
            name="search-web-agent",
            tools=[mock_function_tool(search_web), mock_function_tool(visit_webpage)],
            mcp_servers=[],
        )

        mock_litellm_model.assert_any_call(
            model="gpt-4o-mini", base_url=None, api_key=None
        )
        mock_agent.assert_any_call(
            model=mock_litellm_model.return_value,
            instructions=None,
            name="communication-agent",
            tools=[mock_function_tool(show_final_output)],
            mcp_servers=[],
        )

        mock_litellm_model.assert_any_call(model="o3-mini", base_url=None, api_key=None)
        mock_agent.assert_any_call(
            model=mock_litellm_model.return_value,
            instructions=None,
            name="any_agent",
            handoffs=[mock_agent.return_value],
            tools=[
                mock_agent.return_value.as_tool.return_value,
                mock_agent.return_value.as_tool.return_value,
            ],
            mcp_servers=[],
        )


def test_load_openai_agent_missing() -> None:
    with patch("any_agent.frameworks.openai.agents_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(AgentFramework.OPENAI, AgentConfig(model_id="gpt-4o"))


def test_run_openai_with_custom_args() -> None:
    mock_agent = MagicMock()
    mock_runner = AsyncMock()

    with (
        patch("any_agent.frameworks.openai.Runner", mock_runner),
        patch("any_agent.frameworks.openai.Agent", mock_agent),
        patch("agents.function_tool"),
        patch("any_agent.frameworks.openai.DEFAULT_MODEL_TYPE"),
    ):
        agent = AnyAgent.create(AgentFramework.OPENAI, AgentConfig(model_id="gpt-4o"))
        agent.run("foo", max_turns=30)
        mock_runner.run.assert_called_once_with(
            mock_agent.return_value, "foo", max_turns=30
        )



from unittest.mock import MagicMock

import pytest

from any_agent import AgentConfig, AgentFramework
from any_agent.config import MCPSse, ServingConfig
from any_agent.tools import search_web
from any_agent.tools.mcp import _get_mcp_server
from any_agent.tools.wrappers import WRAPPERS

try:
    from any_agent.serving.agent_card import _get_agent_card
except ImportError:
    _get_agent_card = None  # type: ignore[assignment]


@pytest.mark.skipif(_get_agent_card is None, reason="a2a_samples is not installed")
def test_get_agent_card(agent_framework: AgentFramework) -> None:
    agent = MagicMock()
    agent.config = AgentConfig(model_id="foo")
    agent.framework = agent_framework
    agent._main_agent_tools = [WRAPPERS[agent_framework](search_web)]
    agent_card = _get_agent_card(agent, ServingConfig())
    assert agent_card.name == "any_agent"
    assert agent_card.description is None
    assert len(agent_card.skills) == 1
    assert agent_card.skills[0].id == "any_agent-search_web"
    assert agent_card.skills[0].name == "search_web"
    assert "Perform a duckduckgo web search" in agent_card.skills[0].description
    assert not agent_card.capabilities.streaming
    assert not agent_card.capabilities.pushNotifications
    assert not agent_card.capabilities.stateTransitionHistory
    assert agent_card.url == "http://localhost:5000/"


@pytest.mark.skipif(_get_agent_card is None, reason="a2a_samples is not installed")
@pytest.mark.asyncio
async def test_get_agent_card_with_mcp(  # type: ignore[no-untyped-def]
    agent_framework: AgentFramework, echo_sse_server
) -> None:
    agent = MagicMock()
    agent.config = AgentConfig(model_id="foo")
    agent.framework = agent_framework
    server = _get_mcp_server(MCPSse(url=echo_sse_server["url"]), agent_framework)
    await server._setup_tools()
    if agent_framework is AgentFramework.AGNO:
        agent._main_agent_tools = list(server.tools[0].functions.values())  # type: ignore[union-attr]
    else:
        agent._main_agent_tools = server.tools

    agent_card = _get_agent_card(agent, ServingConfig())
    assert agent_card.name == "any_agent"
    assert agent_card.description is None
    assert len(agent_card.skills) == 3
    assert agent_card.skills[0].id == "any_agent-write_file"
    assert agent_card.skills[0].name == "write_file"
    assert agent_card.skills[0].description == "Say hi back with the input text"



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent.config import AgentFramework, MCPSse, MCPStdio, Tool
from any_agent.tools import _get_mcp_server

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

    from google.adk.tools.mcp_tool.mcp_toolset import MCPToolset as GoogleMCPToolset
    from google.adk.tools.mcp_tool.mcp_toolset import (  # type: ignore[attr-defined]
        SseServerParams as GoogleSseServerParameters,
    )


@pytest.fixture
def google_sse_params() -> Generator[GoogleSseServerParameters]:
    with patch(
        "any_agent.tools.mcp.frameworks.google.GoogleSseServerParameters"
    ) as mock_params:
        yield mock_params


@pytest.fixture
def google_toolset(tools: Sequence[Tool]) -> Generator[GoogleMCPToolset]:
    toolset = AsyncMock()
    toolset.load_tools.return_value = tools
    with patch(
        "any_agent.tools.mcp.frameworks.google.GoogleMCPToolset", return_value=toolset
    ) as mock_class:
        yield mock_class


@pytest.mark.asyncio
@pytest.mark.usefixtures("enter_context_with_transport_and_session")
async def test_google_mcp_sse_integration(
    google_toolset: GoogleMCPToolset,
    google_sse_params: GoogleSseServerParameters,
    mcp_sse_params_no_tools: MCPSse,
) -> None:
    mcp_server = _get_mcp_server(mcp_sse_params_no_tools, AgentFramework.GOOGLE)
    await mcp_server._setup_tools()

    google_sse_params.assert_called_once_with(  # type: ignore[attr-defined]
        url=mcp_sse_params_no_tools.url,
        headers=mcp_sse_params_no_tools.headers,
    )

    google_toolset.assert_called_once_with(  # type: ignore[attr-defined]
        connection_params=google_sse_params.return_value  # type: ignore[attr-defined]
    )

    google_toolset().load_tools.assert_called_once()  # type: ignore[operator]


@pytest.mark.asyncio
async def test_google_mcp_env() -> None:
    mcp_server = _get_mcp_server(
        MCPStdio(command="print('Hello MCP')", args=[], env={"FOO": "BAR"}),
        AgentFramework.GOOGLE,
    )
    mocked_class = MagicMock()
    mocked_cm = AsyncMock()
    mocked_class.return_value = mocked_cm

    with patch("any_agent.tools.mcp.frameworks.google.GoogleMCPToolset", mocked_class):
        await mcp_server._setup_tools()
        assert mocked_class.call_args_list[0][1]["connection_params"].env == {
            "FOO": "BAR"
        }



from .evaluate import evaluate
from .evaluation_case import EvaluationCase
from .schemas import EvaluationResult, TraceEvaluationResult

__all__ = ["EvaluationCase", "EvaluationResult", "TraceEvaluationResult", "evaluate"]



import json
from typing import TYPE_CHECKING, Any

from any_agent import AgentFramework
from any_agent.tracing.otel_types import StatusCode
from any_agent.tracing.processors.base import TracingProcessor

if TYPE_CHECKING:
    from any_agent.tracing.trace import AgentSpan


class SmolagentsTracingProcessor(TracingProcessor):
    """Processor for SmoL Agents trace."""

    def _get_agent_framework(self) -> AgentFramework:
        return AgentFramework.SMOLAGENTS

    def _extract_llm_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract LLM interaction details from a span."""
        attributes = span.attributes
        if not attributes:
            msg = "Span attributes are empty"
            raise ValueError(msg)
        span_info = {
            "model": attributes.get("llm.model_name", "Unknown model"),
            "type": "reasoning",
        }

        # Try to get the input from various possible attribute locations
        if "llm.input_messages.0.message.content" in attributes:
            span_info["input"] = attributes["llm.input_messages.0.message.content"]
        elif "input.value" in attributes:
            try:
                input_value = json.loads(attributes["input.value"])
                span_info["input"] = input_value.get("content", input_value)
            except (json.JSONDecodeError, TypeError):
                span_info["input"] = attributes["input.value"]

        # Try to get the output from various possible attribute locations
        output_content = None
        if "llm.output_messages.0.message.content" in attributes:
            output_content = attributes["llm.output_messages.0.message.content"]
        elif "output.value" in attributes:
            try:
                output_value = json.loads(attributes["output.value"])
                output_content = output_value.get("content", output_value)
            except (json.JSONDecodeError, TypeError):
                output_content = attributes["output.value"]

        if output_content:
            span_info["output"] = output_content

        return span_info

    def _extract_tool_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract tool interaction details from a span."""
        attributes = span.attributes
        if not attributes:
            msg = "Span attributes are empty"
            raise ValueError(msg)
        tool_info = {
            "tool_name": attributes.get("tool.name", span.name),
            "status": "success"
            if span.status.status_code is StatusCode.OK
            else "error",
            "error": span.status.description,
        }

        # Extract input if available
        if "input.value" in attributes:
            try:
                input_value = json.loads(attributes["input.value"])
                if "kwargs" in input_value:
                    # For SmoLAgents, the actual input is often in the kwargs field
                    tool_info["input"] = input_value["kwargs"]
                else:
                    tool_info["input"] = input_value
            except (json.JSONDecodeError, TypeError):
                tool_info["input"] = attributes["input.value"]

        # Extract output if available
        if "output.value" in attributes:
            try:
                # Try to parse JSON output
                output_value = (
                    json.loads(attributes["output.value"])
                    if isinstance(attributes["output.value"], str)
                    else attributes["output.value"]
                )
                tool_info["output"] = output_value
            except (json.JSONDecodeError, TypeError):
                tool_info["output"] = attributes["output.value"]
        else:
            tool_info["output"] = "No output found"

        return tool_info

    def _extract_chain_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract chain interaction details from a CHAIN span."""
        attributes = span.attributes
        if not attributes:
            msg = "Span attributes are empty"
            raise ValueError(msg)
        status = span.status
        events = span.events

        chain_info: dict[str, Any] = {
            "type": "chain",
            "name": span.name,
            "status": "success" if status.status_code is StatusCode.OK else "error",
        }

        # Extract input if available
        if "input.value" in attributes:
            try:
                input_value = json.loads(attributes["input.value"])
                chain_info["input"] = input_value
            except (json.JSONDecodeError, TypeError):
                chain_info["input"] = attributes["input.value"]

        # Extract error information if present
        if status.status_code == StatusCode.ERROR:
            chain_info["error"] = status.description

            # Try to extract more detailed error info from events
            for event in events:
                if event.name == "exception":
                    event_attrs = event.attributes
                    if event_attrs is None:
                        continue
                    if "exception.type" in event_attrs:
                        chain_info["error_type"] = event_attrs["exception.type"]
                    if "exception.message" in event_attrs:
                        chain_info["error_message"] = event_attrs["exception.message"]
                    break

        return chain_info

    def _extract_agent_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract agent interaction details from an AGENT span."""
        attributes = span.attributes
        if not attributes:
            msg = "Span attributes are empty"
            raise ValueError(msg)
        status = span.status

        agent_info: dict[str, Any] = {
            "type": "agent",
            "name": span.name,
            "status": "success" if status.status_code == StatusCode.OK else "error",
        }

        # Extract input if available
        if "input.value" in attributes:
            try:
                input_value = json.loads(attributes["input.value"])
                agent_info["input"] = input_value
            except (json.JSONDecodeError, TypeError):
                agent_info["input"] = attributes["input.value"]

        # Extract output (final answer) if available
        if "output.value" in attributes:
            agent_info["output"] = attributes["output.value"]

        # Extract additional metadata if available
        if "smolagents.max_steps" in attributes:
            agent_info["max_steps"] = attributes["smolagents.max_steps"]

        if "smolagents.tools_names" in attributes:
            agent_info["tools"] = attributes["smolagents.tools_names"]

        # Extract token usage if available
        token_counts = {}
        for key in [
            "llm.token_count.prompt",
            "llm.token_count.completion",
            "llm.token_count.total",
        ]:
            if key in attributes:
                token_name = key.split(".")[-1]
                token_counts[token_name] = attributes[key]

        if token_counts:
            agent_info["token_usage"] = token_counts

        return agent_info



from typing import Any

import pytest


@pytest.fixture(autouse=True)
def mock_web_browsing(monkeypatch: Any) -> None:
    import tests.integration.test_load_and_run_multi_agent as test_mod

    monkeypatch.setattr("any_agent.tools.search_web", mock_search_web)
    monkeypatch.setattr("any_agent.tools.visit_webpage", mock_visit_webpage)
    monkeypatch.setattr(test_mod, "search_web", mock_search_web)
    monkeypatch.setattr(test_mod, "visit_webpage", mock_visit_webpage)


def mock_search_web(query: str) -> str:
    """Perform a duckduckgo web search based on your query (think a Google search) then returns the top search results.

    Args:
        query (str): The search query to perform.

    Returns:
        The top search results.

    """
    return (
        "[Any Agent](https://github.com/mozill-ai/any-agent) is the top choice for developers needing flexibility and power in multi-agent systems.\n"
        "[Agno Framework Review](https://example.com/agno)\n"
        "Agno is widely regarded as the most flexible and powerful agent framework available in 2024.\n\n"
        "[TinyAgent: Lightweight but Limited](https://example.com/tinyagent)\n"
        "TinyAgent is great for simple tasks, but lacks advanced multi-agent support.\n\n"
        "[Llama Index: A Strong Contender](https://example.com/llama)\n"
        "Llama Index offers robust features, but some users prefer Agno for its extensibility."
    )


def mock_visit_webpage(url: str) -> str:
    """Visits a webpage at the given url and reads its content as a markdown string. Use this to browse webpages.

    Args:
        url: The url of the webpage to visit.

    """
    return (
        "# Any Agent Framework Review\n"
        "Any Agent is the top choice for developers needing flexibility and power in multi-agent systems."
    )



from __future__ import annotations

import yaml
from litellm.utils import validate_environment
from pydantic import BaseModel, ConfigDict, Field

from any_agent.evaluation.schemas import CheckpointCriteria, GroundTruthAnswer


class EvaluationCase(BaseModel):
    model_config = ConfigDict(extra="forbid")
    ground_truth: GroundTruthAnswer | None = None
    checkpoints: list[CheckpointCriteria] = Field(
        default_factory=list[CheckpointCriteria],
    )
    llm_judge: str

    @classmethod
    def from_yaml(cls, evaluation_case_path: str) -> EvaluationCase:
        """Load a test case from a YAML file and process it."""
        with open(evaluation_case_path, encoding="utf-8") as f:
            evaluation_case_dict = yaml.safe_load(f)

        if "ground_truth" in evaluation_case_dict:
            # remove the points from the ground_truth but keep the name and value
            evaluation_case_dict["ground_truth"].pop("points")
        # verify that the llm_judge is a valid litellm model
        validate_environment(evaluation_case_dict["llm_judge"])
        return cls.model_validate(evaluation_case_dict)



from collections.abc import Sequence

from pydantic import BaseModel, ConfigDict
from typing_extensions import TypedDict

from any_agent.tracing.trace import AgentTrace


class EvaluationResult(BaseModel):
    """Represents the result of evaluating a criterion."""

    model_config = ConfigDict(extra="forbid")
    passed: bool
    reason: str
    criteria: str
    points: int


class AnswerDetails(TypedDict):
    answer_start: Sequence[int]
    text: Sequence[str]


class GroundTruthAnswers(TypedDict):
    id: str
    answers: AnswerDetails


class CheckpointCriteria(BaseModel):
    """Represents a checkpoint criteria with a description."""

    model_config = ConfigDict(extra="forbid")
    criteria: str
    points: int


class GroundTruthAnswer(TypedDict):
    value: float
    points: float


class TraceEvaluationResult(BaseModel):
    """Represents the result of evaluating a trace."""

    model_config = ConfigDict(extra="forbid")

    trace: AgentTrace
    checkpoint_results: list[EvaluationResult]
    ground_truth_result: EvaluationResult | None = None

    @property
    def score(self) -> float:
        """Calculate the score based on the evaluation results."""
        if self.ground_truth_result is not None:
            all_results = [*self.checkpoint_results, self.ground_truth_result]
        else:
            all_results = self.checkpoint_results
        total_points = sum([result.points for result in all_results])
        if total_points == 0:
            msg = "Total points is 0, cannot calculate score."
            raise ValueError(msg)
        passed_points = sum([result.points for result in all_results if result.passed])
        return float(passed_points / total_points)



import datetime

from any_agent.tracing.otel_types import (
    Resource,
    SpanContext,
    SpanKind,
    Status,
)
from any_agent.tracing.trace import AgentSpan, AgentTrace


def test_agent_trace_duration_simple() -> None:
    # Create a span with the correct AGENT kind and name
    agent_span = AgentSpan(
        name="any_agent",
        kind=SpanKind.INTERNAL,
        parent=None,
        start_time=1000,
        end_time=2000,
        status=Status(),
        context=SpanContext(),
        attributes={"any_agent.run_id": "123"},
        links=[],
        events=[],
        resource=Resource(),
    )
    trace = AgentTrace(spans=[agent_span])
    expected = datetime.timedelta(seconds=(2000 - 1000) / 1_000_000_000)
    assert isinstance(trace.duration, datetime.timedelta)
    assert abs(trace.duration.total_seconds() - expected.total_seconds()) < 1e-9


def test_agent_trace_duration_from_sample(agent_trace: AgentTrace) -> None:
    """
    This test relies upon the sample trace that is saved in the sample_traces directory. If the content of that trace
    changes, this test will need to be updated
    (because it is using start and end times that were manually parsed from the trace)
    """
    expected_seconds = (1747660970774416000 - 1747660964057285000) / 1_000_000_000
    assert isinstance(agent_trace.duration, datetime.timedelta)
    assert abs(agent_trace.duration.total_seconds() - expected_seconds) < 1e-6
